/*
 * nasd_edrfs_worker.c
 *
 * Actual dispatch functions for NASD EDRFS server.
 *
 * Author: Jim Zelenka
 */
/*
 * NASD is a copyright (c) of Carnegie Mellon University, 1997,1998,1999.
 *             
 * Permission to reproduce, use and prepare derivative works of
 * NASD for use is granted provided the copyright and "No Warranty"
 * statements are included with all reproductions and derivative
 * works. NASD may also be redistributed provided that the
 * copyright and "No Warranty" statements are included in all
 * redistributions.
 *             
 * NO WARRANTY. NASD IS FURNISHED ON AN "AS IS" BASIS. CARNEGIE
 * MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
 * LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY,
 * EXCLUSIVITY OF RESULTS OR RESULTS OBTAINED FROM USE OF NASD.
 * CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY
 * KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK OR
 * COPYRIGHT INFRINGEMENT.
 */



#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_edrfs_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_edrfs_internal.h>
#include <nasd/nasd_edrfs_server.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_control.h>
#include <nasd/nasd_timer.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_edrfs_rpc.h>
#include <nasd/nasd_timeout.h>

#if defined(DEC_OSF) || (defined(OSF) && defined(MACH))
#include <sys/secdefines.h>
#include <sys/access.h>
#if SEC_BASE > 0
#include <sys/security.h>
#include <sys/audit.h>
#include <sys/secpolicy.h>
#endif /* SEC_BASE */
#endif /* DEC_OSF || (OSF && MACH) */

#ifdef LINUX
/* grn. */
#define R_OK    4
#define W_OK    2
#define X_OK    1
#define F_OK    0
#endif /* LINUX */

#define WARN_BAD_MARKERV             0
#define NAME_DEBUG                   0

#define DEBUG_RPC_ERR_DETAIL         0

#define DEBUG_READDIR_DETAIL         0
#define DEBUG_LOOKUP_DETAIL          0
#define DEBUG_INTERNAL_LOOKUP_DETAIL 0
#define DEBUG_CREATE_DETAIL          0
#define DEBUG_REMOVE_DETAIL          0
#define DEBUG_RENAME_DETAIL          0
#define DEBUG_SETATTR_DETAIL         0

#define LOOKUP_FORCE_VALID_ATTR      0

#if 0
#define NASD_EDRFS_DATA_BYTE_START 8192
#define NASD_EDRFS_MAX_DATALEN     65536
#else
#define NASD_EDRFS_DATA_BYTE_START 65536
#define NASD_EDRFS_MAX_DATALEN     1310720
#endif

#define OWNER_MODE_BITS(_a_) (_a_ >> 6)
#define GROUP_MODE_BITS(_a_) (_a_ >> 3)
#define OTHER_MODE_BITS(_a_) (_a_)
#define OWNER(x) (((nasd_edrfs_attributes_t *)(x)->fs_specific)->uid)
#define GROUP(x) (((nasd_edrfs_attributes_t *)(x)->fs_specific)->gid)

void
nasd_dump_edrfs_attributes(
  nasd_edrfs_attributes_t  *edrfsattrp)
{
  nasd_printf("type ");
  switch(edrfsattrp->type) {
    case NASD_EDRFS_TYPE_REG:
      nasd_printf("NASD_EDRFS_TYPE_REG ");
      break;
    case NASD_EDRFS_TYPE_DIR:
      nasd_printf("NASD_EDRFS_TYPE_DIR ");
      break;
    case NASD_EDRFS_TYPE_LNK:
      nasd_printf("NASD_EDRFS_TYPE_LNK ");
      break;
    default:
      nasd_printf("0x%x ", ((int)edrfsattrp->type)&0xff);
      break;
  }
  nasd_printf("mode 0x%x nlink %d uid %d gid %d",
    edrfsattrp->mode, edrfsattrp->nlink, edrfsattrp->uid,
    edrfsattrp->gid);
}

/*
 * Actual handlers here
 */

void
nasd_edrfs_real_mount(
  nasd_edrfs_credential_t   in_credential,
  char                     *dirpath,
  nasd_list_len_t           in_listlen,
  nasd_net_address_t        out_drivelist[],
  nasd_list_len_t          *out_listlen,
  nasd_cookie_t            *out_cookie,
  nasd_edrfs_identifier_t  *out_identifier,
  nasd_status_t            *nasd_status)
{
  /* XXX check credential */

  *nasd_status = nasd_edrfs_mount_lookup((char *)dirpath, in_listlen, out_drivelist,
    out_listlen, out_identifier);

  /* @@@ build cookie */
  bzero((char *)out_cookie, sizeof(nasd_cookie_t));
}

void
nasd_edrfs_real_fsstat(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_identifier,
  nasd_edrfs_credential_t       in_credential,
  nasd_edrfs_fsstat_t          *out_stat,    
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  nasd_edrfs_infopage_t *edrfs_info_page;
  nasd_ctrl_part_info_t *ptinfo;
  nasd_edrfs_attr_t *dh_attr;
  nasd_edrfs_core_dir_t *cd;
  nasd_security_param_t sp;
  nasd_timespec_t ts;
  nasd_status_t rc;
  nasd_timer_t tm;
  nasd_uint64 n;
  int partnum;

  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }
  edrfs_info_page = nasd_edrfs_get_infopage();
  if (edrfs_info_page == NULL) {
    *nasd_status = NASD_NO_MEM;
    return;
  }

  NASD_EDRFS_LOCK_DIRCACHE();

  rc = nasd_edrfs_dir_hash_lookup(in_identifier, 1, &cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(dh_attr);
    *nasd_status = rc;
    NASD_EDRFS_UNLOCK_DIRCACHE();
  }
  else {
    NASD_EDRFS_UNLOCK_DIRCACHE();
    NASD_ASSERT(cd);
    partnum = cd->drive->partnum;
    ptinfo = (nasd_ctrl_part_info_t *)edrfs_info_page->info_page;
    /* @@@ check in_cookie, build cookie */
    bzero(&sp, sizeof(nasd_security_param_t));
    sp.actual_protection = NASD_NO_PROTECTION;
    NASD_TM_START(&tm);
    rc = nasd_cl_p_ctrl_get_part_info(cd->drive->handle, NULL, &sp, NULL,
      partnum, ptinfo);
    NASD_TM_STOP(&tm);
    NASD_TM_ELAPSED_TS(&tm, &ts);
    NASD_ATOMIC_TIMESPEC_ADD(&nasd_edrfs_cache_stats.drive_time, &ts);
    if (rc) {
      nasd_printf("EDRFS: nasd_cl_p_get_part_info: got 0x%x (%s)\n",
        rc, nasd_error_string(rc));
      *nasd_status = rc;
    }
    else {
      if (out_stat) {
        out_stat->total_bytes = ptinfo->part_size * cd->drive->blocksize;
        n = ptinfo->part_size - ptinfo->blocks_allocated;
        out_stat->free_bytes = n * cd->drive->blocksize;
        out_stat->user_free_bytes = out_stat->free_bytes;
        out_stat->total_files = ptinfo->max_objs;
        out_stat->free_files = ptinfo->max_objs - ptinfo->num_obj;
        out_stat->user_free_files = out_stat->free_files;
        out_stat->invariant_seconds = 0; /* this is what EDRFS does too */
      }
      *nasd_status = NASD_SUCCESS;
    }
    NASD_EDRFS_LOCK_DIRCACHE();
    nasd_edrfs_dir_release_post_attribute(cd, in_identifier, post_attribute);
    NASD_EDRFS_UNLOCK_DIRCACHE();
  }

  if (edrfs_info_page) {
    nasd_edrfs_free_infopage(edrfs_info_page);
  }
}

void
nasd_edrfs_real_fsinfo(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_identifier,
  nasd_edrfs_credential_t       in_credential,
  nasd_edrfs_fsinfo_t          *out_info,    
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }

  if (out_info) {
    out_info->max_read_size             = NASD_EDRFS_MAX_DATALEN;
    out_info->preferred_read_size       = NASD_EDRFS_DATA_BYTE_START;
    out_info->preferred_read_multiple   = NASD_EDRFS_DATA_BYTE_START;
    out_info->max_write_size            = NASD_EDRFS_MAX_DATALEN;
    out_info->preferred_write_size      = NASD_EDRFS_DATA_BYTE_START;
    out_info->preferred_write_multiple  = NASD_EDRFS_DATA_BYTE_START;
    out_info->preferred_readdir_size    = NASD_EDRFS_DATA_BYTE_START;
    out_info->max_file_size             = INT_MAX;
    out_info->time_delta.ts_sec         = 0;
    out_info->time_delta.ts_nsec        = 10000000;
    out_info->fs_properties             = 0; /* might change if further support added */
  }

  if (post_attribute) {
    nasd_edrfs_lookup_post_attribute(in_identifier, post_attribute);
  }

  *nasd_status = NASD_SUCCESS;
}

void
nasd_edrfs_real_lookup(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_identifier,
  nasd_edrfs_credential_t       in_credential,
  char                         *in_dirpath,
  nasd_cookie_t                *out_cookie,
  nasd_edrfs_identifier_t      *out_identifier,
  nasd_edrfs_post_attribute_t  *out_attribute,
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  nasd_edrfs_diskdirent_t *de;
  nasd_edrfs_attr_t *dh_attr;
  nasd_edrfs_core_dir_t *cd;
  nasd_status_t rc;

#if LOOKUP_FORCE_VALID_ATTR > 0
  nasd_p_getattr_dr_res_t ga_res;
#endif /* LOOKUP_FORCE_VALID_ATTR > 0 */

#if DEBUG_INTERNAL_LOOKUP_DETAIL > 0
  nasd_printf("internal_lookup: starting for nasdid 0x%" NASD_ID_FMT
         " in_dirpath=\"%s\"\n", in_identifier.nasd_identifier,
         in_dirpath);
#endif /* DEBUG_INTERNAL_LOOKUP_DETAIL > 0 */

  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }
  if (out_attribute) {
    out_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }

  NASD_EDRFS_LOCK_DIRCACHE();
  rc = nasd_edrfs_dir_hash_lookup(in_identifier, 1, &cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(dh_attr);
  }
  else {
    NASD_ASSERT(cd);
    NASD_EDRFS_LOCK_DIR(cd,1);
    /* XXX check permissions */
    de = nasd_edrfs_name_lookup(cd, (char *)in_dirpath);
    if (de == NULL) {
      rc = NASD_EDRFS_BAD_NAME;
    }
    else {
      out_identifier->nasd_identifier = de->data->nasdid;
      out_identifier->disk_identifier = cd->di;
      out_identifier->partnum=cd->drive->partnum;

      if (out_attribute) {
        if (de->attr) {
          bcopy((char *)&NASD_EDRFS_ATTR_ATTR(de->attr),
                (char *)&out_attribute->attribute,
                sizeof(nasd_attribute_t));
          out_attribute->valid = NASD_EDRFS_ATTRIBUTE_VALID;
          rc = NASD_SUCCESS;
        }
        else {
#if LOOKUP_FORCE_VALID_ATTR > 0
          NASD_EDRFS_CSINC(lookup_get_attr);
          rc = nasd_edrfs_do_getattr(cd->drive, de->data->nasdid, &ga_res);
          if (rc == NASD_SUCCESS) {
            bcopy((char *)&ga_res.out_attribute,
              (char *)&out_attribute->attribute,
              sizeof(nasd_attribute_t));
            out_attribute->valid = NASD_EDRFS_ATTRIBUTE_VALID;
          }
#else /* LOOKUP_FORCE_VALID_ATTR > 0 */
          out_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
          rc = NASD_SUCCESS;
#endif /* LOOKUP_FORCE_VALID_ATTR > 0 */
        }
      }
      else {
        rc = NASD_SUCCESS;
      }
    }
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release_post_attribute(cd, in_identifier, post_attribute);
  }

  NASD_EDRFS_UNLOCK_DIRCACHE();

  if (rc == NASD_SUCCESS) {
    /* @@@ build cookie */
    bzero((char *)out_cookie, sizeof(nasd_cookie_t));
  }

#if DEBUG_INTERNAL_LOOKUP_DETAIL > 0
  nasd_printf("internal_lookup: finished, returning with status 0x%x (%s)\n",
         rc, nasd_error_string(rc));
#endif /* DEBUG_INTERNAL_LOOKUP_DETAIL > 0 */

  *nasd_status = rc;
}

void
nasd_edrfs_real_readdir(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_identifier,
  nasd_edrfs_credential_t       in_credential,
  nasd_edrfs_marker_t           in_marker,    
  nasd_edrfs_markerv_t          in_markerv,
  nasd_len_t                    in_count,
  nasd_edrfs_marker_t          *out_marker,
  nasd_edrfs_markerv_t         *out_markerv,
  nasd_len_t                   *out_count,
  nasd_boolean_t               *out_eof,
  nasd_procpipe_t              *byte_pipe,
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  nasd_edrfs_dirent_otw_t entry_otw;
  nasd_edrfs_core_dirpage_t *cdp;
  nasd_edrfs_attr_t *dh_attr;
  int i, pageind, ui, ub, pi;
  nasd_edrfs_marker_t marker;
  nasd_edrfs_dirent_t entry;
  nasd_edrfs_markerv_t verf;
  nasd_edrfs_core_dir_t *cd;
  nasd_edrfs_dirdata_t *dd;
  nasd_status_t rc;

#if DEBUG_READDIR_DETAIL > 0
  nasd_printf("readdir_fm called on nasdid 0x%" NASD_ID_FMT
         " in_marker 0x%x in_markerv 0x%x in_count %d\n",
         in_identifier.nasd_identifier, in_marker, in_markerv,
         in_count);
#endif /* DEBUG_READDIR_DETAIL > 0 */

  rc = NASD_SUCCESS;

  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }

  *out_count = 0;

  NASD_EDRFS_LOCK_DIRCACHE();

  rc = nasd_edrfs_dir_hash_lookup(in_identifier, 1, &cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(dh_attr);
    *nasd_status = rc;
  }
  else {
    NASD_ASSERT(cd);
    /* XXX check credential, permissions */
    NASD_EDRFS_LOCK_DIR(cd,1);
    rc = nasd_edrfs_dir_markerv(cd, &verf);
    if (rc) {
      *nasd_status = rc;
      goto done_inner_readdir;
    }
    if (in_marker && (in_markerv != verf)) {
      if(in_markerv != 0) {
        *nasd_status = rc = NASD_EDRFS_BAD_MARKER_VERF;
        goto done_inner_readdir;
      } else {
        /*
         * Why look, it's a little lost DEC EDRFS client.
         * It wants to talk to us, but it's too cool
         * for RFC 1813. I guess we'll talk to it.
         */
#if WARN_BAD_MARKERV > 0
        nasd_printf("readdir_fm WARNING: request with in_marker!=0 and in_markerv==0 received\n");
        nasd_printf("readdir_fm WARNING: perhaps your client is from DEC? ignoring bad markerv\n");
#endif /* WARN_BAD_MARKERV */
      }
    }
    cdp = cd->first_page;
    pageind = 0;
    marker = 0;

    while(cdp&&((*out_count)<in_count)) {
      ui = pageind/8;
      ub = 1<<(pageind%8);
      if (cdp->page->header.usemap[ui]&ub) {
        /* valid entry */
        marker++;
        pi = 1;
        if (marker >= in_marker) {
          dd = &cdp->page->data_slots[pageind];
          /*
           * Send the entry to the client
           */
#if DEBUG_READDIR_DETAIL > 0
          nasd_printf("readdir_fm returning entry \"%s\" nasdid 0x%" NASD_ID_FMT
                 "\n", dd->name, dd->nasdid);
#endif /* DEBUG_READDIR_DETAIL > 0 */

          entry.nd_ino = dd->nasdid;
          entry.nd_namlen = dd->name_len;
          memcpy(entry.nd_name, dd->name, dd->name_len);
          entry.nd_name[dd->name_len] = '\0';
          nasd_edrfs_dirent_t_marshall(&entry, entry_otw);
          rc = byte_pipe->push(byte_pipe->state, entry_otw, sizeof(entry_otw),
                               NULL, NULL, NULL);
          if (rc) {
            break;
          }
          i++;
          pi = NASD_EDRFS_DIRD_EXLEN(dd)+1;
          (*out_count)++; /* total # dirents sent */
        }
        else {
          pi = 1;
        }
      }
      else {
        pi = 1;
      }
      pageind += pi;
      if (pageind >= 127) {
        NASD_ASSERT(pageind == 127);
        pageind = 0;
        cdp = cdp->next;
      }
    }

#if DEBUG_READDIR_DETAIL > 0
    nasd_printf("readdir_fm out_count=%d in_count=%d\n", *out_count, in_count);
#endif /* DEBUG_READDIR_DETAIL > 0 */

    NASD_ASSERT(*out_count <= in_count);
    if (cdp) {
      /* there's more to dir */
      *out_eof = 0;
      *out_marker = marker + 1;
    }
    else {
      *out_eof = 1;
      /*
       * The RFC (1813) suggests that we can (and implies should) set
       * out_marker to 0 here. However, if we do so, the client
       * misbehaves (in the layer of NFS code on top of the NASD
       * EDRFS dispatcher). I'm not sure if this is a bogusity
       * in everyone's implementation of NFS, or if I'm misunderstanding
       * the RFC. --jimz
       */
      *out_marker = marker + 1;
    }
    *out_markerv = verf;
done_inner_readdir:
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release_post_attribute(cd, in_identifier, post_attribute);
  }
  *nasd_status = rc;

  NASD_EDRFS_UNLOCK_DIRCACHE();

#if DEBUG_READDIR_DETAIL > 0
  nasd_printf("readdir_fm called on nasdid 0x%" NASD_ID_FMT
         " in_count %d result 0x%x (%s)\n",
         in_identifier.nasd_identifier, in_count, *nasd_status,
         nasd_error_string(*nasd_status));
  nasd_printf("readdir_fm returning out_marker 0x%" NASD_64x_FMT
         " out_markerv 0x%" NASD_64x_FMT " out_count %d out_eof %d\n",
         *out_marker, *out_markerv, *out_count, *out_eof);
#endif /* DEBUG_READDIR_DETAIL > 0 */
}

void
nasd_edrfs_real_access(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_identifier,
  nasd_edrfs_credential_t       in_credential,
  nasd_edrfs_access_t           in_access,
  nasd_edrfs_access_t          *out_access,
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  nasd_edrfs_attributes_t edrfsattr;
  nasd_p_getattr_dr_res_t ga_res;
  nasd_edrfs_attr_t *dh_attr;
  nasd_edrfs_core_dir_t *cd;
  nasd_attribute_t *attr;
  nasd_identifier_t nid;
  nasd_edrfs_drive_t *dr;
  nasd_status_t rc;
  int need_doit;

  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }
  if (out_access == NULL) {
    *nasd_status = NASD_FAIL;
    return;
  }

  *out_access = 0;
  need_doit = 0;
  nid = NASDID_NULL;

  NASD_EDRFS_LOCK_DIRCACHE();

  cd = NULL;
  rc = nasd_edrfs_dir_hash_lookup(in_identifier, 1, &cd, &dh_attr);
  if (rc) {
    cd = NULL;
    NASD_EDRFS_CSINC(access_lookup_fail);
    if (rc == NASD_EDRFS_NOT_DIR) {
      /*
       * The object is valid, but not a directory.
       * Perform access checks, but then ditch the
       * attribute.
       */
      NASD_EDRFS_CSINC(access_not_dir);
      attr = &dh_attr->attr_res.out_attribute;
      edrfsattr = *(nasd_edrfs_attributes_t *)attr->fs_specific;
      *nasd_status = NASD_SUCCESS;
      nasd_edrfs_release_attr(dh_attr);
      dh_attr = NULL;
      goto other_attr;
    }
    *nasd_status = rc;
  }
  else {
    NASD_ASSERT(dh_attr == NULL);
    if (cd) {
      NASD_EDRFS_LOCK_DIR(cd,1);
      rc = NASD_SUCCESS;
      if (cd->attr == NULL) {
        NASD_EDRFS_CSINC(access_fault_attr);
        rc = nasd_edrfs_dir_load_attr(cd);
      }
      if (rc == NASD_SUCCESS) {
        attr = &NASD_EDRFS_ATTR_ATTR(cd->attr);
        edrfsattr = *(nasd_edrfs_attributes_t *)attr->fs_specific;
      }
      NASD_EDRFS_UNLOCK_DIR(cd);
    }
    else {
      rc = nasd_edrfs_get_drive(in_identifier, &dr);
      if (rc == NASD_SUCCESS) {
        need_doit = 1;
        nid = cd->nid;
      }
    }
    *nasd_status = rc;
  }

  if (cd) {
    nasd_edrfs_dir_release_post_attribute(cd, in_identifier, post_attribute);
  }
other_attr:
  NASD_EDRFS_UNLOCK_DIRCACHE();

  if (((*nasd_status) == NASD_SUCCESS) && need_doit) {
    NASD_ASSERT(dr);
    NASD_EDRFS_CSINC(access_get_attr);
    rc = nasd_edrfs_do_getattr(dr, nid, &ga_res);
    if (rc == NASD_SUCCESS) {
      attr = &ga_res.out_attribute;
      edrfsattr = *(nasd_edrfs_attributes_t *)attr->fs_specific;
    }
    *nasd_status = rc;
  }

  if ((*nasd_status) == NASD_SUCCESS) {
    /*
     * Check access on the file
     * XXX Is this right? This is what the old NFS server did,
     * but it seems to be missing group, all privs
     */
    if (in_access & NASD_EDRFS_ACCESS_READ) {
      if (OWNER_MODE_BITS(edrfsattr.mode) & R_OK) {
        *out_access |= NASD_EDRFS_ACCESS_READ;
      }
    }
    if ((in_access & NASD_EDRFS_ACCESS_LOOKUP)
      || (in_access & NASD_EDRFS_ACCESS_EXECUTE))
    {
      if (OWNER_MODE_BITS(edrfsattr.mode) & X_OK) {
        *out_access |= NASD_EDRFS_ACCESS_LOOKUP;
        *out_access |= NASD_EDRFS_ACCESS_EXECUTE;
      }
    }
    if ((in_access & NASD_EDRFS_ACCESS_MODIFY)
      || (in_access & NASD_EDRFS_ACCESS_EXTEND)
      || (in_access & NASD_EDRFS_ACCESS_DELETE))
    {
      if (OWNER_MODE_BITS(edrfsattr.mode) & W_OK) {
        *out_access |= NASD_EDRFS_ACCESS_MODIFY;
        *out_access |= NASD_EDRFS_ACCESS_EXTEND;
        *out_access |= NASD_EDRFS_ACCESS_DELETE;
      }
    }
  }
}

void
nasd_edrfs_real_setattr(
  nasd_cookie_t             in_cookie,
  nasd_edrfs_identifier_t   in_identifier,
  nasd_edrfs_credential_t   in_credential,
  nasd_attribute_t          in_attribute,
  nasd_fieldmask_t          in_fieldmask,
  nasd_guard_t              in_guard,
  nasd_attribute_t         *out_attribute,
  nasd_status_t            *nasd_status)
{
  nasd_edrfs_attributes_t *edrfsattr, *new_edrfsattr;
  nasd_attribute_t *attr, *old_attr, *new_attr;
  nasd_p_setattr_dr_args_t *sa_args;
  nasd_fieldmask_t edrfs_fieldmask;
  nasd_p_setattr_dr_res_t *sa_res;
  nasd_p_getattr_dr_res_t *ga_res;
  nasd_edrfs_setattrpair_t *sapair;
  nasd_edrfs_attr_t *dh_attr;
  nasd_edrfs_core_dir_t *cd;
  nasd_edrfs_drive_t *dr;
  nasd_status_t rc;

  *nasd_status = NASD_SUCCESS;

#if DEBUG_SETATTR_DETAIL > 0
  nasd_printf("setattr_fm: setattr starting for nasdid 0x%" NASD_ID_FMT
         " uid=%d gid=%d fieldmask=0x%x guard=0x%x attributes=\"",
         in_identifier.nasd_identifier,
         in_credential.uid, in_credential.gid, in_fieldmask, in_guard);
  nasd_dump_edrfs_attributes((nasd_edrfs_attributes_t *)in_attribute.fs_specific);
  nasd_printf("\"\n");
#endif /* DEBUG_SETATTR_DETAIL > 0 */

  sapair = nasd_edrfs_get_setattrpair();
  if (sapair == NULL) {
    *nasd_status = NASD_NO_MEM;
    goto really_done_setattr;
  }
  sa_args = &sapair->sa_args;
  sa_res = &sapair->sa_res;
  ga_res = &sapair->ga_res;

  NASD_EDRFS_LOCK_DIRCACHE();

  rc = nasd_edrfs_dir_hash_lookup(in_identifier, 0, &cd, &dh_attr);
  if (rc) {
    NASD_ASSERT(rc != NASD_EDRFS_NOT_DIR);
    cd = NULL;
    *nasd_status = rc;
    goto done_setattr;
  }
  NASD_ASSERT(dh_attr == NULL);
  if (cd) {
    NASD_EDRFS_LOCK_DIR(cd,1);
  }
  NASD_EDRFS_UNLOCK_DIRCACHE();

  if (cd) {
    if (cd->attr == NULL) {
      rc = nasd_edrfs_dir_load_attr(cd);
      if (rc) {
        *nasd_status = rc;
        goto really_done_setattr;
      }
    }
    attr = &NASD_EDRFS_ATTR_ATTR(cd->attr);
    dr = cd->drive;
  }
  else {
    rc = nasd_edrfs_get_drive(in_identifier, &dr);
    if (rc) {
      *nasd_status = rc;
      goto really_done_setattr;
    }
    NASD_ASSERT(dr != NULL);
    NASD_EDRFS_CSINC(setattr_get_attr);
    rc = nasd_edrfs_do_getattr(dr, in_identifier.nasd_identifier, ga_res);
    if (rc) {
      *nasd_status = rc;
      goto really_done_setattr;
    }
    attr = &ga_res->out_attribute;
  }

  /*
   * Now we have the old object attributes.
   * Figure out the necessary swizzling and set
   * it into sa_args.
   */
  sa_args->in_identifier = in_identifier.nasd_identifier;
  bcopy((char *)attr, (char *)&sa_args->in_attribute, sizeof(nasd_attribute_t));
  sa_args->in_fieldmask = 0;
  sa_args->in_partnum = dr->partnum;
  sa_args->in_guard = (in_guard ? 1 : 0);
  old_attr = attr;
  attr = &sa_args->in_attribute;
  edrfsattr = (nasd_edrfs_attributes_t *)attr->fs_specific;
  new_attr = &in_attribute;
  new_edrfsattr = (nasd_edrfs_attributes_t *)new_attr->fs_specific;
  edrfs_fieldmask = in_fieldmask;

  if (edrfs_fieldmask&NASD_ATTR_FS_SPECIFIC) {
    edrfs_fieldmask &= ~NASD_ATTR_FS_SPECIFIC;
    if (edrfs_fieldmask & (NASD_EDRFS_ATTR_TYPE | NASD_EDRFS_ATTR_NLINK)) {
      *nasd_status = NASD_AUTHCHECK_FAILED;
      goto done_setattr_nolock;
    }
    if (edrfs_fieldmask & NASD_EDRFS_ATTR_MODE) {
      edrfs_fieldmask &= ~NASD_EDRFS_ATTR_MODE;
      if ((in_credential.uid == edrfsattr->uid)
        || (in_credential.uid == NASD_EDRFS_ROOT_UID))
      {
        edrfsattr->mode = new_edrfsattr->mode;
      }
      else {
        *nasd_status = NASD_AUTHCHECK_FAILED;
        goto done_setattr_nolock;
      }
    }
    if (edrfs_fieldmask & NASD_EDRFS_ATTR_UID) {
      edrfs_fieldmask &= ~NASD_EDRFS_ATTR_UID;
      if ((in_credential.uid == edrfsattr->uid)
        || (in_credential.uid == NASD_EDRFS_ROOT_UID))
      {
        edrfsattr->uid = new_edrfsattr->uid;
      }
      else {
        *nasd_status = NASD_AUTHCHECK_FAILED;
        goto done_setattr_nolock;
      }
    }
    if (edrfs_fieldmask & NASD_EDRFS_ATTR_GID) {
      edrfs_fieldmask &= ~NASD_EDRFS_ATTR_GID;
      if ((in_credential.uid == edrfsattr->uid)
        || (in_credential.uid == NASD_EDRFS_ROOT_UID))
      {
        edrfsattr->gid = new_edrfsattr->gid;
      }
      else {
        *nasd_status = NASD_AUTHCHECK_FAILED;
        goto done_setattr_nolock;
      }
    }
    sa_args->in_fieldmask |= NASD_ATTR_FS_SPECIFIC;
  }
  if (edrfs_fieldmask&(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
    |NASD_ATTR_ATTR_MODIFY_TIME))
  {
    edrfs_fieldmask &= ~(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
      |NASD_ATTR_ATTR_MODIFY_TIME);
    if ((in_credential.uid == edrfsattr->uid)
      || (in_credential.uid ==NASD_EDRFS_ROOT_UID))
    {
      /* drive will filter for us */
      sa_args->in_fieldmask |=
        in_fieldmask&(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
        |NASD_ATTR_ATTR_MODIFY_TIME);
      attr->object_len = in_attribute.object_len;
      attr->fs_attr_modify_time = in_attribute.fs_attr_modify_time;
      attr->fs_object_modify_time = in_attribute.fs_object_modify_time;
    }
    else {
      *nasd_status = NASD_AUTHCHECK_FAILED;
      goto done_setattr_nolock;
    }
  }
  if (edrfs_fieldmask) {
    *nasd_status = NASD_BAD_FIELDMASK;
    goto done_setattr_nolock;
  }
  /*
   * Now're we're ready to set some attributes.
   */
  sapair->otw_buf.args_otw = sapair->sa_args_otw;
  sapair->otw_buf.res_otw = sapair->sa_res_otw;
  rc = nasd_edrfs_do_setattr(dr, sa_args, sa_res, &sapair->otw_buf);
  if (rc == NASD_SUCCESS) {
    if (cd) {
      bcopy((char *)&sa_res->out_attribute,
        (char *)&NASD_EDRFS_ATTR_ATTR(cd->attr), sizeof(nasd_attribute_t));
    }
  }
  bcopy((char *)&sa_res->out_attribute, (char *)out_attribute,
    sizeof(nasd_attribute_t));

  *nasd_status = rc;

done_setattr_nolock:
  if (cd) {
    /* patch to correct locking prototol */
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release(cd);
    cd = NULL;
    goto really_done_setattr;
  }
  NASD_EDRFS_LOCK_DIRCACHE();

done_setattr:
  if (cd) {
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release(cd);
  }
  NASD_EDRFS_UNLOCK_DIRCACHE();
really_done_setattr:
  if (sapair) {
    nasd_edrfs_free_setattrpair(sapair);
  }

#if DEBUG_SETATTR_DETAIL > 0
  nasd_printf("setattr_fm: setattr returning 0x%x (%s), attributes: ",
         *nasd_status, nasd_error_string(*nasd_status));
  nasd_dump_edrfs_attributes((nasd_edrfs_attributes_t *)out_attribute->fs_specific);
  nasd_printf("\n");
#endif /* DEBUG_SETATTR_DETAIL > 0 */

}

void
nasd_edrfs_real_create(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_directory,
  nasd_edrfs_credential_t       in_credential,
  char                         *in_dirpath,
  nasd_attribute_t              in_attribute,
  nasd_fieldmask_t              in_fieldmask,
  nasd_cookie_t                *out_cookie,
  nasd_edrfs_identifier_t      *out_identifier,
  nasd_attribute_t             *out_attribute,
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  nasd_edrfs_createpair_t *crpair;
  nasd_edrfs_diskdirent_t *de;
  nasd_status_t rc;
  int lk;

  crpair = NULL;
  lk = 0;
  de = NULL;

  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }

  rc = nasd_edrfs_valid_name((char *)in_dirpath);
  if (rc) {
    *nasd_status = rc;
    goto really_done_create;
  }

  crpair = nasd_edrfs_get_createpair();
  if (crpair == NULL) {
    *nasd_status = NASD_NO_MEM;
    goto really_done_create;
  }

  NASD_EDRFS_LOCK_DIRCACHE();
  lk = 1;
  rc = nasd_edrfs_dir_hash_lookup(in_directory, 1, &crpair->cd,
    &crpair->dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(crpair->dh_attr);
    *nasd_status = rc;
    goto done_create;
  }

  NASD_ASSERT(crpair->cd);
  NASD_EDRFS_LOCK_DIR(crpair->cd,1);

  /* XXX check permissions */

#if 0 /* we'll catch this later */
  de = nasd_edrfs_name_lookup(cd, (char *)in_dirpath);
  if (de) {
    *nasd_status = NASD_EDRFS_ALREADY_EXISTS;
    goto done_create2;
  }
#endif

  NASD_EDRFS_UNLOCK_DIRCACHE();
  lk = 0;

  /*
   * We've established that no entry is in this directory
   * with the given name. Create a new object on the drive
   * to be this file.
   */

  /* @@@ init cookie [not sure if we need to do that anymore here -mju] */
  crpair->attr = &crpair->args.in_attribute;
  crpair->edrfsattr = (nasd_edrfs_attributes_t *)crpair->attr->fs_specific;
  crpair->new_edrfsattr = (nasd_edrfs_attributes_t *)in_attribute.fs_specific;
  crpair->args.in_fieldmask = NASD_ATTR_FS_SPECIFIC;

  crpair->edrfsattr->type = NASD_EDRFS_TYPE_REG;
  crpair->edrfsattr->mode = 0100000;
  crpair->edrfsattr->nlink = 1;
  crpair->edrfsattr->uid = in_credential.uid;
  crpair->edrfsattr->gid = in_credential.gid;
  crpair->edrfsattr->clean = NASD_EDRFS_CLEAN;
#if NAME_DEBUG > 0
  crpair->i = NASD_MIN(NASD_EDRFS_NAME_DEBUG_LEN-1, strlen(in_dirpath));
  NASD_ASSERT(crpair->i > 0);
  bcopy((char *)in_dirpath, (char *)crpair->edrfsattr->name_debug, crpair->i);
  crpair->edrfsattr->name_debug[crpair->i] = '\0';
#else /* NAME_DEBUG > 0 */
  crpair->edrfsattr->name_debug[0] = '\0';
#endif /* NAME_DEBUG > 0 */
  crpair->fieldmask = in_fieldmask;

  if (crpair->fieldmask&NASD_ATTR_FS_SPECIFIC) {
    crpair->fieldmask &= ~NASD_ATTR_FS_SPECIFIC;
    if (crpair->fieldmask & NASD_EDRFS_ATTR_TYPE) {
      crpair->fieldmask &= ~NASD_EDRFS_ATTR_TYPE;
      if (crpair->new_edrfsattr->type != NASD_EDRFS_TYPE_REG) {
        /* currently not supported */
        *nasd_status = NASD_OP_NOT_SUPPORTED;
        goto done_create2;
      }
    }
    if (crpair->fieldmask & NASD_EDRFS_ATTR_NLINK) {
      crpair->fieldmask &= ~NASD_EDRFS_ATTR_NLINK;
      *nasd_status = NASD_AUTHCHECK_FAILED;
      goto done_create2;
    }
    if (crpair->fieldmask & NASD_EDRFS_ATTR_MODE) {
      crpair->fieldmask &= ~NASD_EDRFS_ATTR_MODE;
      crpair->edrfsattr->mode = crpair->new_edrfsattr->mode;
    }
    if (crpair->fieldmask & NASD_EDRFS_ATTR_UID) {
      crpair->fieldmask &= ~NASD_EDRFS_ATTR_UID;
      crpair->edrfsattr->uid = crpair->new_edrfsattr->uid;
    }
    if (crpair->fieldmask & NASD_EDRFS_ATTR_GID) {
      crpair->fieldmask &= ~NASD_EDRFS_ATTR_GID;
      crpair->edrfsattr->gid = crpair->new_edrfsattr->gid;
    }
  }
  if (crpair->fieldmask&(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
    |NASD_ATTR_ATTR_MODIFY_TIME))
  {
    crpair->fieldmask &= ~(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
      |NASD_ATTR_ATTR_MODIFY_TIME);
    /* drive will filter for us */
    crpair->args.in_fieldmask |=
      crpair->fieldmask&(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
      |NASD_ATTR_ATTR_MODIFY_TIME);
    crpair->attr->object_len = in_attribute.object_len;
    crpair->attr->fs_attr_modify_time = in_attribute.fs_attr_modify_time;
    crpair->attr->fs_object_modify_time = in_attribute.fs_object_modify_time;
  }
  if (crpair->fieldmask) {
    *nasd_status = NASD_BAD_FIELDMASK;
    goto done_create2;
  }
  crpair->args.in_fieldmask |= NASD_ATTR_AV|NASD_ATTR_LAYOUT_HINT;
  crpair->args.in_attribute.av = NASD_EDRFS_INITIAL_AV;
  crpair->args.in_attribute.layout_hint.lh_nid = crpair->cd->nid;
  crpair->otw_buf.args_otw = crpair->args_otw;
  crpair->otw_buf.res_otw = crpair->res_otw;
  rc = nasd_edrfs_do_create(crpair->cd->drive, &crpair->args, &crpair->res,
                          &crpair->otw_buf);
  if (rc) {
    *nasd_status = rc;
    goto done_create2;
  }

  NASD_EDRFS_UNLOCK_DIR(crpair->cd);
  NASD_EDRFS_LOCK_DIRCACHE();
  lk = 1;
  NASD_EDRFS_LOCK_DIR(crpair->cd,1);

  /*
   * Have to re-lookup here because someone else might
   * have created it in the meantime.
   */
  de = nasd_edrfs_name_lookup(crpair->cd, (char *)in_dirpath);
  if (de) {
    rc = nasd_edrfs_do_remove(crpair->cd->drive, crpair->res.out_identifier);
    if (rc) {
      NASD_PANIC();
    }
    *nasd_status = NASD_EDRFS_ALREADY_EXISTS;
    goto done_create2;
  }
  /*
   * Now the object exists, so we can add it to the directory.
   */
  crpair->flags = NASD_EDRFS_DIRD_TYPE_DATA<<NASD_EDRFS_DIRD_SHIFT;
  rc = nasd_edrfs_dir_add_entry(crpair->cd, (char *)in_dirpath,
    crpair->res.out_identifier, crpair->flags, &de);
  if (rc) {
    rc = nasd_edrfs_do_remove(crpair->cd->drive, crpair->res.out_identifier);
    if (rc) {
      NASD_PANIC();
    }
    *nasd_status = NASD_FAIL;
    goto done_create2;
  }

  /* XXX increment nlink here! */


  NASD_EDRFS_UNLOCK_DIRCACHE();
  lk = 0;
  /* @@@ generate cookie */
  bzero((char *)out_cookie, sizeof(nasd_cookie_t));
  out_identifier->nasd_identifier = crpair->res.out_identifier;
  out_identifier->disk_identifier = crpair->cd->drive->di;
  out_identifier->partnum = crpair->cd->drive->partnum;
 
  *out_attribute = crpair->res.out_attribute;
  *nasd_status = NASD_SUCCESS;

done_create2:
  NASD_EDRFS_UNLOCK_DIR(crpair->cd);
  if (lk == 0) {
    NASD_EDRFS_LOCK_DIRCACHE();
    lk = 1;
  }
  nasd_edrfs_dir_release_post_attribute(crpair->cd, *out_identifier,
    post_attribute);

done_create:
  if (lk) {
    NASD_EDRFS_UNLOCK_DIRCACHE();
    lk = 0;
  }

really_done_create:

  if (crpair) {
    nasd_edrfs_free_createpair(crpair);
  }
  NASD_ASSERT(lk == 0);

#if DEBUG_CREATE_DETAIL > 0
  if (*nasd_status == 0) {
    nasd_printf("create returning success creating \"%s\" in directory with nasdid 0x%"
      NASD_ID_FMT " new file has id 0x%" NASD_ID_FMT "\n",
      in_dirpath, in_directory.nasd_identifier, out_identifier->nasd_identifier);
    if (de) {
      nasd_printf("create of \"%s\" got de 0x%lx nasdid 0x%" NASD_ID_FMT " pagename \"%s\" len %d\n",
        in_dirpath, de, de->data->nasdid, de->data->name, de->data->name_len);
    }
  }
#endif /* DEBUG_CREATE_DETAIL > 0 */
}

void
nasd_edrfs_real_symlink(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_directory,
  nasd_edrfs_credential_t       in_credential,
  char                         *in_dirpath,
  char                         *in_symlink,
  nasd_attribute_t              in_attribute,
  nasd_cookie_t                *out_cookie,
  nasd_edrfs_identifier_t      *out_identifier,
  nasd_attribute_t             *out_attribute,
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  nasd_edrfs_attributes_t *edrfsattr, *new_edrfsattr;
  nasd_p_smpl_op_dr_args_t p_write_args;
  nasd_p_fastwrite_dr_res_t p_write_res;
  nasd_security_param_t sp;
  nasd_edrfs_createpair_t *crpair;
  nasd_edrfs_diskdirent_t *de;
  nasd_edrfs_attr_t *dh_attr;
  nasd_rpc_status_t wrstat;
  nasd_edrfs_core_dir_t *cd;
  nasd_attribute_t *attr;
  nasd_timespec_t ts;
  nasd_uint64 flags;
  nasd_status_t rc;
  nasd_timer_t tm;
  int lk;

#if NAME_DEBUG > 0
  int i;
#endif /* NAME_DEBUG > 0 */

  crpair = NULL;
  lk = 0;
  de = NULL;

  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }

  rc = nasd_edrfs_valid_name((char *)in_dirpath);
  if (rc) {
    *nasd_status = rc;
    goto really_done_symlink;
  }

  if (strlen((const char *) in_symlink) >= NASD_EDRFS_MAX_NAME_LEN) {
    *nasd_status = NASD_EDRFS_BAD_NAME;
    goto really_done_symlink;
  }

  crpair = nasd_edrfs_get_createpair();
  if (crpair == NULL) {
        *nasd_status = NASD_NO_MEM;
    goto really_done_symlink;
  }
  
  NASD_EDRFS_LOCK_DIRCACHE();
  lk = 1;
  rc = nasd_edrfs_dir_hash_lookup(in_directory, 1, &cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR) { nasd_edrfs_release_attr(dh_attr); }
    *nasd_status = rc;
    goto done_symlink;
  }

  NASD_ASSERT(cd);
  NASD_EDRFS_LOCK_DIR(cd,1);

  /* XXX check permissions */

  NASD_EDRFS_UNLOCK_DIRCACHE();
  lk = 0;

  /* @@@ init capability */
  attr = &crpair->args.in_attribute;
  edrfsattr = (nasd_edrfs_attributes_t *)attr->fs_specific;
  new_edrfsattr = (nasd_edrfs_attributes_t *)in_attribute.fs_specific;
  crpair->args.in_fieldmask = NASD_ATTR_FS_SPECIFIC;

  edrfsattr->type = NASD_EDRFS_TYPE_LNK;
  edrfsattr->mode = 0777;
  edrfsattr->nlink = 1;
  edrfsattr->uid = in_credential.uid;
  edrfsattr->gid = in_credential.gid;
  edrfsattr->clean = NASD_EDRFS_CLEAN;
#if NAME_DEBUG > 0
  i = NASD_MIN(NASD_EDRFS_NAME_DEBUG_LEN-1, strlen(in_dirpath));
  NASD_ASSERT(i > 0);
  bcopy((char *)in_dirpath, (char *)edrfsattr->name_debug, i);
  edrfsattr->name_debug[i] = '\0';
#else /* NAME_DEBUG > 0 */
  edrfsattr->name_debug[0] = '\0';
#endif /* NAME_DEBUG > 0 */

  crpair->args.in_fieldmask |= NASD_ATTR_AV|NASD_ATTR_LAYOUT_HINT;
  crpair->args.in_attribute.av = NASD_EDRFS_INITIAL_AV;
  crpair->args.in_attribute.layout_hint.lh_nid = cd->nid;
  crpair->otw_buf.args_otw = crpair->args_otw;
  crpair->otw_buf.res_otw = crpair->res_otw;
  rc = nasd_edrfs_do_create(cd->drive, &crpair->args, &crpair->res,
                          &crpair->otw_buf);
  if (rc) {
    *nasd_status = NASD_FAIL;
    goto done_symlink2;
  }

  NASD_EDRFS_UNLOCK_DIR(cd);
  NASD_EDRFS_LOCK_DIRCACHE();
  lk = 1;
  NASD_EDRFS_LOCK_DIR(cd,1);

  de = nasd_edrfs_name_lookup(cd, (char *)in_dirpath);
  if (de) {
    rc = nasd_edrfs_do_remove(cd->drive, crpair->res.out_identifier);
    if (rc) {
      NASD_PANIC();
    }
    *nasd_status = NASD_EDRFS_ALREADY_EXISTS;
    goto done_symlink2;
  }

  flags = NASD_EDRFS_DIRD_TYPE_LINK<<NASD_EDRFS_DIRD_SHIFT;
  rc = nasd_edrfs_dir_add_entry(cd, (char *)in_dirpath, crpair->res.out_identifier,
                              flags, &de);
  if (rc) {
    rc = nasd_edrfs_do_remove(cd->drive, crpair->res.out_identifier);
    if (rc) {
      NASD_PANIC();
    }
    *nasd_status = NASD_FAIL;
    goto done_symlink2;
  }

  NASD_EDRFS_UNLOCK_DIRCACHE();
  lk = 0;

  /* we've created the object. now we need to write to it. */

  sp.type = in_cookie.capability.type;
  sp.partnum = cd->drive->partnum;
  sp.actual_protection = NASD_NO_PROTECTION;

  p_write_args.in_identifier = crpair->res.out_identifier;
  p_write_args.in_offset = 0;
  p_write_args.in_len = strlen((char *) in_symlink);
  p_write_args.in_partnum = cd->drive->partnum;

  NASD_TM_START(&tm);
  nasd_cl_p_write_simple_dr(cd->drive->handle,
                            in_cookie.key, &sp, &in_cookie.capability,
                            &p_write_args,
                            (void *) in_symlink,
                            &p_write_res, &wrstat);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_edrfs_cache_stats.drive_time, &ts);

  rc = p_write_res.nasd_status;
  if (rc) {
    nasd_printf("uh-oh. had a problem writing: '%s'\n", nasd_error_string(rc));
  }
  if (wrstat) {
    rc = NASD_FAIL;
  }

  bzero((char *)out_cookie, sizeof(nasd_cookie_t));
  out_identifier->nasd_identifier = crpair->res.out_identifier;
  out_identifier->disk_identifier = cd->drive->di;
  out_identifier->partnum = cd->drive->partnum;
 
  *out_attribute = crpair->res.out_attribute;
  *nasd_status = rc;

 done_symlink2:
  NASD_EDRFS_UNLOCK_DIR(cd);
  if (lk == 0) {
    NASD_EDRFS_LOCK_DIRCACHE();
    lk = 1;
  }
  nasd_edrfs_dir_release_post_attribute(cd, *out_identifier, post_attribute);

 done_symlink:
  if (lk) {
    NASD_EDRFS_UNLOCK_DIRCACHE();
    lk = 0;
  }

 really_done_symlink:

  if (crpair) {
    nasd_edrfs_free_createpair(crpair);
  }
  NASD_ASSERT(lk == 0);

#if DEBUG_SYMLINK_DETAIL > 0
  if (*nasd_status == 0) {
    nasd_printf("returning success creating link \"%s\" (nasdid 0x%"
           NASD_ID_FMT ") pointing to \"%s\"\nin directory with nasdid 0x%"
           NASD_ID_FMT "\n",
           in_dirpath, out_identifier->nasd_identifier, in_symlink,
           in_directory.nasd_identifier);
    if (de) {
      nasd_printf("create of \"%s\" got de 0x%lx nasdid 0x%"
             NASD_ID_FMT " pagename \"%s\" len %d\n",
             in_dirpath, de, de->data->nasdid,
             de->data->name, de->data->name_len);
    }
  }
#endif /* DEBUG_SYMLINK_DETAIL > 0 */
}

void
nasd_edrfs_real_remove(
  nasd_cookie_t             in_cookie,
  nasd_edrfs_identifier_t   in_directory,
  nasd_edrfs_credential_t   in_credential,
  char                     *in_dirpath,
  nasd_status_t            *nasd_status)
{
  nasd_edrfs_diskdirent_t *de;
  nasd_identifier_t nasdid;
  nasd_edrfs_attr_t *dh_attr;
  nasd_edrfs_core_dir_t *cd;
  nasd_status_t rc;

  NASD_EDRFS_LOCK_DIRCACHE();
  rc = nasd_edrfs_dir_hash_lookup(in_directory, 1, &cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(dh_attr);
    *nasd_status = rc;
    NASD_EDRFS_UNLOCK_DIRCACHE();
    return;
  }

  NASD_ASSERT(cd);
  NASD_ASSERT(dh_attr == NULL);

  NASD_EDRFS_LOCK_DIR(cd, 1);

  de = nasd_edrfs_name_lookup(cd, (char *)in_dirpath);

  if (de == NULL) {
    *nasd_status = NASD_EDRFS_BAD_NAME;
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release(cd);
    NASD_EDRFS_UNLOCK_DIRCACHE();
    return;
  }

#if DEBUG_REMOVE_DETAIL > 0
  nasd_printf("remove found ent 0x%lx name \"%s\" for \"%s\" in 0x%"
          NASD_ID_FMT " dataname \"%s\" len %d\n",
          de, de->name, in_dirpath, in_directory.nasd_identifier,
          de->data->name, de->data->name_len);
#endif /* DEBUG_REMOVE_DETAIL > 0 */

  if (NASD_EDRFS_DIRD_TYPEOF(de->data) == NASD_EDRFS_DIRD_TYPE_DIR) {
    *nasd_status = NASD_EDRFS_IS_DIR;
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release(cd);
    NASD_EDRFS_UNLOCK_DIRCACHE();
    return;
  }

  nasdid = de->data->nasdid;

  /*
   * Remove entry from dir
   */
  rc = nasd_edrfs_dir_remove_entry(cd, de);
  if (rc) {
    NASD_PANIC();
  }
  NASD_EDRFS_UNLOCK_DIR(cd);

  nasd_edrfs_dir_release(cd);
  NASD_EDRFS_UNLOCK_DIRCACHE();

  /*
   * Remove entry from drive
   */
  rc = nasd_edrfs_do_remove(cd->drive, nasdid);
  if (rc) {
    nasd_printf("EDRFS: failed to remove id 0x%" NASD_ID_FMT " error 0x%x (%s)\n",
      nasdid, rc, nasd_error_string(rc));
    nasd_printf("     id was for \"%s\" in directory with ID 0x%" NASD_ID_FMT "\n",
      in_dirpath, in_directory.nasd_identifier);
    NASD_PANIC();
    *nasd_status = NASD_FAIL;
  }

  *nasd_status = NASD_SUCCESS;

#if DEBUG_REMOVE_DETAIL > 0
  if (*nasd_status == 0) {
    nasd_printf("remove returning success deleting \"%s\" in directory with nasdid 0x%"
      NASD_ID_FMT "\n",
      in_dirpath, in_directory.nasd_identifier);
  }
#endif /* DEBUG_REMOVE_DETAIL > 0 */
}

void
nasd_edrfs_real_mkdir(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_directory,
  nasd_edrfs_credential_t       in_credential,
  char                         *in_dirpath,
  nasd_attribute_t              in_attribute,
  nasd_fieldmask_t              in_fieldmask,
  nasd_cookie_t                *out_cookie,
  nasd_edrfs_identifier_t      *out_identifier,
  nasd_attribute_t             *out_attribute,
  nasd_edrfs_post_attribute_t  *post_attribute,
  nasd_status_t                *nasd_status)
{
  nasd_edrfs_diskdirent_t *de, *dde_dot, *dde_dot_dot;
  nasd_edrfs_attributes_t *edrfsattr, *new_edrfsattr;
  nasd_edrfs_createpair_t *crpair;
  nasd_edrfs_core_dir_t *cd, *new;
  nasd_edrfs_dirpage_t *dirpage;
  nasd_fieldmask_t fieldmask;
  nasd_edrfs_attr_t *dh_attr;
  nasd_attribute_t *attr;
  int dir_valid, i, lk;
  nasd_uint64 flags;
  nasd_status_t rc;

  crpair = NULL;
  lk = 0;

  if (post_attribute) {
    post_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }

  rc = nasd_edrfs_valid_name((char *)in_dirpath);
  if (rc) {
    *nasd_status = rc;
    goto really_done_mkdir;
  }

  crpair = nasd_edrfs_get_createpair();
  if (crpair == NULL) {
    *nasd_status = NASD_NO_MEM;
    goto really_done_mkdir;
  }

  NASD_EDRFS_LOCK_DIRCACHE();
  lk = 1;

  cd = NULL;
  dir_valid = 0;

  /* get a dir structure for this new directory */
  new = nasd_edrfs_dir_get_coredir(1);
  if (new == NULL) {
    *nasd_status = NASD_NO_MEM;
    goto done_mkdir;
  }
  new->attr = nasd_edrfs_get_attr();

  /* now get a page to hold the data */
  nasd_edrfs_dir_mk_freepages(1);
  new->first_page = new->last_page = nasd_edrfs_free_dirpages;
  nasd_edrfs_free_dirpages = new->last_page->next;
  nasd_edrfs_num_free_dirpages--;
  new->npages = 1;
  new->first_page->prev = NULL;
  new->last_page->next = NULL;
  new->first_page->offset = 0;
  dirpage = new->first_page->page;

  /* get dir to put entry into */
  rc = nasd_edrfs_dir_hash_lookup(in_directory, 1, &cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(dh_attr);
    *nasd_status = rc;
    goto done_mkdir;
  }

  NASD_ASSERT(cd);

  new->drive = cd->drive;
  new->di = cd->drive->di;
  new->isroot = 0;

  NASD_EDRFS_LOCK_DIR(cd,1);

  /*
   * XXX check permissions
   */

#if 0 /* we'll catch this later */
  de = nasd_edrfs_name_lookup(cd, (char *)in_dirpath);
  if (de) {
    *nasd_status = NASD_EDRFS_ALREADY_EXISTS;
    goto done_mkdir2;
  }
#endif

  NASD_EDRFS_UNLOCK_DIRCACHE();
  lk = 0;

  /*
   * We've established that no entry is in this directory
   * with the given name. Create a new object on the drive
   * to be this directory.
   */

  /* @@@ init cookie -- not sure we need to do this here anymore -mju */
  attr = &crpair->args.in_attribute;
  edrfsattr = (nasd_edrfs_attributes_t *)attr->fs_specific;
  new_edrfsattr = (nasd_edrfs_attributes_t *)in_attribute.fs_specific;
  crpair->args.in_fieldmask = NASD_ATTR_FS_SPECIFIC;

  edrfsattr->type = NASD_EDRFS_TYPE_DIR;
  edrfsattr->mode = 0040000;
  edrfsattr->nlink = 2;
  edrfsattr->uid = in_credential.uid;
  edrfsattr->gid = in_credential.gid;
  edrfsattr->clean = NASD_EDRFS_CLEAN;
#if NAME_DEBUG > 0
  i = NASD_MIN(NASD_EDRFS_NAME_DEBUG_LEN-1, strlen(in_dirpath));
  NASD_ASSERT(i > 0);
  bcopy((char *)in_dirpath, (char *)edrfsattr->name_debug, i);
  edrfsattr->name_debug[i] = '\0';
#else /* NAME_DEBUG > 0 */
  edrfsattr->name_debug[0] = '\0';
#endif /* NAME_DEBUG > 0 */
  fieldmask = in_fieldmask;

  if (fieldmask&NASD_ATTR_FS_SPECIFIC) {
    fieldmask &= ~NASD_ATTR_FS_SPECIFIC;
    if (fieldmask & NASD_EDRFS_ATTR_TYPE) {
      fieldmask &= ~NASD_EDRFS_ATTR_TYPE;
      if (new_edrfsattr->type != NASD_EDRFS_TYPE_DIR) {
        /* currently not supported */
        *nasd_status = NASD_OP_NOT_SUPPORTED;
        goto done_mkdir2;
      }
    }
    if (fieldmask & NASD_EDRFS_ATTR_NLINK) {
      fieldmask &= ~NASD_EDRFS_ATTR_NLINK;
      *nasd_status = NASD_AUTHCHECK_FAILED;
      goto done_mkdir2;
    }
    if (fieldmask & NASD_EDRFS_ATTR_MODE) {
      fieldmask &= ~NASD_EDRFS_ATTR_MODE;
      edrfsattr->mode = new_edrfsattr->mode;
    }
    if (fieldmask & NASD_EDRFS_ATTR_UID) {
      fieldmask &= ~NASD_EDRFS_ATTR_UID;
      edrfsattr->uid = new_edrfsattr->uid;
    }
    if (fieldmask & NASD_EDRFS_ATTR_GID) {
      fieldmask &= ~NASD_EDRFS_ATTR_GID;
      edrfsattr->gid = new_edrfsattr->gid;
    }
  }
  if (fieldmask&(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
    |NASD_ATTR_ATTR_MODIFY_TIME))
  {
    fieldmask &= ~(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
      |NASD_ATTR_ATTR_MODIFY_TIME);
    /* drive will filter for us */
    crpair->args.in_fieldmask |=
      fieldmask&(NASD_ATTR_OBJECT_LEN|NASD_ATTR_MODIFY_TIME
      |NASD_ATTR_ATTR_MODIFY_TIME);
    attr->object_len = in_attribute.object_len;
    attr->fs_attr_modify_time = in_attribute.fs_attr_modify_time;
    attr->fs_object_modify_time = in_attribute.fs_object_modify_time;
  }
  if (fieldmask) {
    *nasd_status = NASD_BAD_FIELDMASK;
    goto done_mkdir2;
  }
  crpair->args.in_fieldmask |= NASD_ATTR_AV;
  crpair->args.in_attribute.av = NASD_EDRFS_INITIAL_AV;
  crpair->otw_buf.args_otw = crpair->args_otw;
  crpair->otw_buf.res_otw = crpair->res_otw;
  rc = nasd_edrfs_do_create(cd->drive, &crpair->args, &crpair->res,
                          &crpair->otw_buf);
  if (rc) {
    *nasd_status = NASD_FAIL;
    goto done_mkdir2;
  }

  bcopy((char *)&crpair->res.out_attribute,
    (char *)&new->attr->attr_res.out_attribute,
    sizeof(nasd_attribute_t));
  new->attr->attr_res.nasd_status = NASD_SUCCESS;
  new->nid = crpair->res.out_identifier;

  /*
   * The new object exists, turn it into a directory
   * by writing out a directory entry (which contains
   * "." and "..").
   */

  /*
   * Format the page
   */
  dirpage->header.usemap[0] = 3;
  for(i=1;i<16;i++)
    dirpage->header.usemap[i] = 0;
  dirpage->header.freecnt = 125;

  dirpage->data_slots[0].flags = NASD_EDRFS_DIRD_TYPE_DIR<<NASD_EDRFS_DIRD_SHIFT;
  dirpage->data_slots[0].nasdid = crpair->res.out_identifier;
  dirpage->data_slots[0].name_len = 1;
  dirpage->data_slots[0].name[0] = '.';
  dirpage->data_slots[0].name[1] = '\0';

  dirpage->data_slots[1].flags = NASD_EDRFS_DIRD_TYPE_DIR<<NASD_EDRFS_DIRD_SHIFT;
  dirpage->data_slots[1].nasdid = cd->nid;
  dirpage->data_slots[1].name_len = 2;
  dirpage->data_slots[1].name[0] = '.';
  dirpage->data_slots[1].name[1] = '.';
  dirpage->data_slots[1].name[2] = '\0';

  /*
   * Hash . and ..
   */

  /* hash . */
  dde_dot = nasd_edrfs_name_add(new, new->first_page, &dirpage->data_slots[0]);
  if (dde_dot == NULL) {
    NASD_PANIC();
  }
  NASD_ASSERT(dde_dot->attr == NULL);
  dde_dot->attr = new->attr;
  dde_dot->attr->refcnt++;

  /* hash .. */
  dde_dot_dot = nasd_edrfs_name_add(new, new->first_page, &dirpage->data_slots[1]);
  if (dde_dot_dot == NULL) {
    NASD_PANIC();
  }
  NASD_ASSERT(dde_dot_dot->attr == NULL);
  dde_dot_dot->attr = cd->attr;
  dde_dot_dot->attr->refcnt++;

  /*
   * Write the directory
   */
  new->first_page->dirty = 1;
  new->ndirty = 1;
  rc = nasd_edrfs_dir_write_dirty(new);
  if (rc) {
    *nasd_status = NASD_FAIL;
    goto done_mkdir2;
  }

  NASD_EDRFS_UNLOCK_DIR(cd);
  NASD_EDRFS_LOCK_DIRCACHE();
  lk = 1;
  NASD_EDRFS_LOCK_DIR(cd,1);

  /*
   * Have to re-lookup here because someone else might
   * have created it in the meantime.
   */
  de = nasd_edrfs_name_lookup(cd, (char *)in_dirpath);
  if (de) {
    rc = nasd_edrfs_do_remove(cd->drive, crpair->res.out_identifier);
    if (rc) {
      NASD_PANIC();
    }
    *nasd_status = NASD_EDRFS_ALREADY_EXISTS;
    goto done_mkdir2;
  }

  /*
   * Now the directory exists, so we can add it to the directory.
   */
  flags = NASD_EDRFS_DIRD_TYPE_DIR<<NASD_EDRFS_DIRD_SHIFT;
  rc = nasd_edrfs_dir_add_entry(cd, (char *)in_dirpath, crpair->res.out_identifier,
    flags, &de);
  if (rc) {
    rc = nasd_edrfs_do_remove(cd->drive, crpair->res.out_identifier);
    if (rc) {
      NASD_PANIC();
    }
    *nasd_status = NASD_FAIL;
    goto done_mkdir2;
  }
  de->attr = new->attr;
  de->attr->refcnt++;
  NASD_ASSERT(de->attr->refcnt == 3);
  dir_valid = 1;

  /*
   * Add it to the directory hash
   */
  NASD_EDRFS_LOCK_DIR(new,1);
  nasd_edrfs_dir_hash_ins(new);
  NASD_EDRFS_UNLOCK_DIR(new);

  dir_valid = 2;

  /* @@@ generate cookie */
  bzero((char *)out_cookie, sizeof(nasd_cookie_t));
  out_identifier->nasd_identifier = crpair->res.out_identifier;
  out_identifier->disk_identifier = cd->drive->di;
  out_identifier->partnum = cd->drive->partnum;

  *out_attribute = new->attr->attr_res.out_attribute;
  *nasd_status = NASD_SUCCESS;

done_mkdir2:
  NASD_EDRFS_UNLOCK_DIR(cd);

done_mkdir:
  if (dir_valid == 2) {
    nasd_edrfs_dir_release(new);
  }
  if (dir_valid == 1) {
    rc = nasd_edrfs_dir_remove_entry(cd, de);
    if (rc) {
      NASD_PANIC();
    }
    nasd_edrfs_dealloc_coredir(new, 1);
  }
  else if (dir_valid == 0) {
    if (new) {
      nasd_edrfs_dealloc_coredir(new, 1);
    }
  }
  if (cd) {
    nasd_edrfs_dir_release_post_attribute(cd, *out_identifier, post_attribute);
  }

  if (lk) {
    NASD_EDRFS_UNLOCK_DIRCACHE();
    lk = 0;
  }

really_done_mkdir:

  if (crpair) {
    nasd_edrfs_free_createpair(crpair);
  }
  NASD_ASSERT(lk == 0);
}

void
nasd_edrfs_real_rmdir(
  nasd_cookie_t             in_cookie,
  nasd_edrfs_identifier_t   in_directory,
  nasd_edrfs_credential_t   in_credential,
  char                     *in_dirpath,
  nasd_status_t            *nasd_status)
{
  nasd_edrfs_core_dir_t *cd, *rm_dir;
  nasd_edrfs_identifier_t rm_dir_id;
  nasd_edrfs_diskdirent_t *de;
  nasd_identifier_t nasdid;
  nasd_edrfs_attr_t *dh_attr;
  nasd_edrfs_drive_t *drive;
  nasd_status_t rc;

  NASD_EDRFS_LOCK_DIRCACHE();

  rc = nasd_edrfs_dir_hash_lookup(in_directory, 1, &cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(dh_attr);
    *nasd_status = rc;
    NASD_EDRFS_UNLOCK_DIRCACHE();
    return;
  }
  NASD_ASSERT(cd);
  NASD_ASSERT(dh_attr == NULL);

  NASD_EDRFS_LOCK_DIR(cd, 1);

  de = nasd_edrfs_name_lookup(cd, (char *)in_dirpath);

  if (de == NULL) {
    *nasd_status = NASD_EDRFS_BAD_NAME;
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release(cd);
    NASD_EDRFS_UNLOCK_DIRCACHE();
    return;
  }

  if (NASD_EDRFS_DIRD_TYPEOF(de->data) != NASD_EDRFS_DIRD_TYPE_DIR) {
    *nasd_status = NASD_EDRFS_NOT_DIR;
    NASD_EDRFS_UNLOCK_DIR(cd);
    nasd_edrfs_dir_release(cd);
    NASD_EDRFS_UNLOCK_DIRCACHE();
    return;
  }

  nasdid = de->data->nasdid;

  NASD_EDRFS_UNLOCK_DIR(cd);

  /*
   * Load in child dir so we can flag it for deletion
   * (we could probably do something more clever).
   */
  rm_dir_id.nasd_identifier = de->data->nasdid;
  rm_dir_id.disk_identifier = cd->drive->di;
  rc = nasd_edrfs_dir_hash_lookup(rm_dir_id, 1, &rm_dir, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR) {
      /* should never happen */
      nasd_edrfs_release_attr(dh_attr);
      NASD_PANIC();
    }
    nasd_edrfs_dir_release(cd);
    *nasd_status = rc;
    NASD_PANIC();
    return;
  }
  NASD_ASSERT(rm_dir);
  NASD_ASSERT(dh_attr == NULL);

  if (rm_dir->nentries > 2) {
    nasd_edrfs_dir_release(cd);
    nasd_edrfs_dir_release(rm_dir);
    rm_dir = NULL;
    NASD_EDRFS_UNLOCK_DIRCACHE();
    *nasd_status = NASD_EDRFS_DIR_NOT_EMPTY;
    return;
  }

  NASD_EDRFS_LOCK_DIR(rm_dir,1);

  rm_dir->deleted = 1;

  NASD_EDRFS_UNLOCK_DIR(rm_dir);

  NASD_EDRFS_LOCK_DIR(cd,1);
  /*
   * Remove entry from parent dir
   */
  rc = nasd_edrfs_dir_remove_entry(cd, de);
  if (rc) {
    NASD_PANIC();
  }
  NASD_EDRFS_UNLOCK_DIR(cd);

  while(rm_dir->refcnt > 1) {
    NASD_WAIT_COND(rm_dir->cond, nasd_edrfs_dircache_lock);
  }

  /*
   * We no longer need the dir lock on the dir to remove,
   * because no one else is allowed to be touching this
   * dir, since we set deleted above, and gave everyone
   * their chance to get out (and they did). It also no
   * longer has an entry in the parent dir. Release the
   * (dirty) parent dir for rewrite, then we can get
   * rid of locks and remove the object from the drive.
   */

  drive = cd->drive;

  nasd_edrfs_dir_release(cd);
  NASD_EDRFS_UNLOCK_DIRCACHE();

  /*
   * Remove entry from drive
   */
  rc = nasd_edrfs_do_remove(drive, nasdid);
  if (rc) {
    NASD_PANIC();
  }

  NASD_EDRFS_LOCK_DIRCACHE();
  nasd_edrfs_dir_release(rm_dir);
  NASD_EDRFS_UNLOCK_DIRCACHE();

  *nasd_status = NASD_SUCCESS;
}

void
nasd_edrfs_real_newcookie(
  nasd_edrfs_identifier_t   in_identifier,
  nasd_edrfs_credential_t   in_credential,
  nasd_cookie_t            *out_cookie,
  nasd_status_t            *nasd_status)
{
  /* @@@ implement! */
  bzero((char *)out_cookie, sizeof(nasd_cookie_t));
  *nasd_status = NASD_SUCCESS;
}

void
nasd_edrfs_real_rename(
  nasd_cookie_t                 in_cookie,
  nasd_edrfs_identifier_t       in_from_directory,
  nasd_edrfs_credential_t       in_credential,
  char                         *in_from_path,
  nasd_edrfs_identifier_t       in_to_directory,
  char                         *in_to_path,
  nasd_edrfs_post_attribute_t  *post_from_attribute,
  nasd_edrfs_post_attribute_t  *post_to_attribute,
  nasd_status_t                *nasd_status)
{
  int to_len, to_ents_required, from_ents_required, n, ntype, otype;
  nasd_edrfs_diskdirent_t *old_de, *new_de, *ditch_de;
  nasd_edrfs_core_dir_t *from_cd, *to_cd, *new_cd;
  nasd_identifier_t del_nasdid;
  nasd_edrfs_drive_t *del_drive;
  nasd_edrfs_attr_t *dh_attr;
  nasd_uint64 flags;
  nasd_status_t rc;

#if DEBUG_RENAME_DETAIL > 0
  nasd_printf("rename 0x%" NASD_ID_FMT ":%s to 0x%" NASD_ID_FMT ":%s\n",
    in_from_directory.nasd_identifier, in_from_path,
    in_to_directory.nasd_identifier, in_to_path);
#endif /* DEBUG_RENAME_DETAIL > 0 */

  if (post_from_attribute) {
    post_from_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }
  if (post_to_attribute) {
    post_to_attribute->valid = NASD_EDRFS_ATTRIBUTE_INVALID;
  }
  from_cd = NULL;
  to_cd = NULL;
  old_de = NULL;
  new_de = NULL;
  new_cd = NULL;
  del_drive = NULL;
  del_nasdid = NASDID_NULL;
  ditch_de = NULL;
  ntype = NASD_EDRFS_DIRD_TYPE_BOGUS;

  if ((strcmp(in_to_path, ".") == 0) || (strcmp(in_from_path, ".") == 0)) {
    *nasd_status = NASD_EDRFS_BAD_NAME;
    goto really_done_rename;
  }

  to_len = strlen(in_to_path);
  if (to_len > NASD_EDRFS_MAX_NAME_LEN) {
    *nasd_status = NASD_EDRFS_BAD_NAME;
    goto really_done_rename;
  }
  if (to_len < 1) {
    *nasd_status = NASD_EDRFS_BAD_NAME;
    goto really_done_rename;
  }

  to_len++; /* hold the NULL */
  to_ents_required = 1;
  if (to_len > NASD_EDRFS_DIRDATA_EMBEDDED_NAME_SIZE) {
    n = to_len - NASD_EDRFS_DIRDATA_EMBEDDED_NAME_SIZE;
    to_ents_required += (n+NASD_EDRFS_DIRDATA_SIZE-1)/NASD_EDRFS_DIRDATA_SIZE;
  }

  NASD_EDRFS_LOCK_DIRCACHE();
  rc = nasd_edrfs_dir_hash_lookup(in_from_directory, 1, &from_cd, &dh_attr);
  if (rc) {
    if (rc == NASD_EDRFS_NOT_DIR)
      nasd_edrfs_release_attr(dh_attr);
    *nasd_status = rc;
    goto done_rename;
  }

  if ((in_from_directory.nasd_identifier == in_to_directory.nasd_identifier)
    && (in_from_directory.disk_identifier == in_to_directory.disk_identifier)
    && (in_from_directory.partnum == in_to_directory.partnum))
  {
    to_cd = from_cd;
    to_cd->refcnt++; /* can do this because we hold dircache lock */
  }
  else {
    rc = nasd_edrfs_dir_hash_lookup(in_to_directory, 1, &to_cd, &dh_attr);
    if (rc) {
      if (rc == NASD_EDRFS_NOT_DIR)
        nasd_edrfs_release_attr(dh_attr);
      *nasd_status = rc;
      goto done_rename;
    }
  }

  /* XXX check permissions */

  NASD_ASSERT(from_cd);
  NASD_ASSERT(to_cd);

  NASD_EDRFS_LOCK_DIR(from_cd,1);
  if (to_cd != from_cd) {
    NASD_EDRFS_LOCK_DIR(to_cd,1);
  }

  old_de = nasd_edrfs_name_lookup(from_cd, (char *)in_from_path);
  if (old_de == NULL) {
    *nasd_status = NASD_EDRFS_BAD_NAME;
    goto done_rename2;
  }
  new_de = nasd_edrfs_name_lookup(to_cd, (char *)in_to_path);
  if (new_de) {
    /* Check to make sure they are not the same */
    if ((from_cd == to_cd) && (old_de->data->nasdid == new_de->data->nasdid)) {
      *nasd_status = NASD_SUCCESS;
      goto done_rename2; 
    }
    
    ntype = NASD_EDRFS_DIRD_TYPEOF(new_de->data);
    otype = NASD_EDRFS_DIRD_TYPEOF(old_de->data);
    if (((ntype == NASD_EDRFS_DIRD_TYPE_DIR)
         || (otype == NASD_EDRFS_DIRD_TYPE_DIR))
        && (ntype != otype)) {
      /*
       * One is a directory, the other is not
       * client error will end up as EEXIST
       */
      *nasd_status = NASD_EDRFS_ALREADY_EXISTS;
      goto done_rename2;
    }
    if (ntype == NASD_EDRFS_DIRD_TYPE_DIR) {
      /*
       * Both are directories. Is the 'target'
       * directory empty?
       */
      nasd_edrfs_identifier_t tmp; /* Make a directory id for the to_path */
      tmp.nasd_identifier = new_de->data->nasdid;
      tmp.disk_identifier = in_to_directory.disk_identifier;
      tmp.partnum = in_to_directory.partnum;

      rc = nasd_edrfs_dir_hash_lookup(tmp, 1, &new_cd, &dh_attr);
      if (rc) {
        if (rc == NASD_EDRFS_NOT_DIR)
          NASD_PANIC();
        new_cd = NULL;
        *nasd_status = rc;
        goto done_rename2;
      }
      NASD_EDRFS_LOCK_DIR(new_cd,1);
      if (new_cd->nentries > 2) {
        /* dir not empty */
        *nasd_status = NASD_EDRFS_DIR_NOT_EMPTY;
        goto done_rename2;
      }
    }
    ditch_de = new_de;
  }

  /*
   * old_de is the dirent for the source of the rename
   * new_de is the dirent for the target (if it's null,
   *   the target does not exist)
   * from_cd is the directory old_de is in
   * to_cd is the directory new_de is in
   * ditch_de is new_de
   * new_cd is dir that is ditch_de (subdir of to_cd)
   */

  from_ents_required = NASD_EDRFS_DIRD_EXLEN(old_de->data)+1;

  if (ditch_de && (to_cd == from_cd) && (from_ents_required == to_ents_required)) {
    /*
     * Src name pages = dest name pages
     * Dest already exists (ditch_de), and may be a directory (new_cd)
     */

#if DEBUG_RENAME_DETAIL > 0
    nasd_printf("rename old_de    0x%lx name \"%s\" id 0x%"
            NASD_ID_FMT "\n",
            old_de, old_de->name, old_de->data->nasdid);
    if (ditch_de) {
      nasd_printf("rename ditch_de 0x%lx name \"%s\" id 0x%"
              NASD_ID_FMT "\n",
              ditch_de, ditch_de->name, ditch_de->data->nasdid);
    }
#endif /* DEBUG_RENAME_DETAIL > 0 */

    /* get rid of ent for original */
    rc = nasd_edrfs_name_remove(from_cd, old_de, 1);
    if (rc) {
      NASD_PANIC();
    }

    /* set up target ent for new name */
    new_de = old_de;
    nasd_edrfs_dir_dirty_page(to_cd, new_de->page);
    strcpy(new_de->name, in_to_path);
    new_de->data->name_len = to_len - 1; /* take 1 for NULL back off */
    rc = nasd_edrfs_name_hash_ins(to_cd, new_de);
    if (rc) {
      NASD_PANIC();
    }

    del_nasdid = ditch_de->data->nasdid;
    if (new_cd) {
      /*
       * Old target was empty directory, new_cd is its core
       */
      new_cd->deleted = 1;
      NASD_EDRFS_UNLOCK_DIR(new_cd);
      while(new_cd->refcnt > 1) {
        NASD_WAIT_COND(new_cd->cond, nasd_edrfs_dircache_lock);
      }
      del_drive = new_cd->drive;
    }
    else {
      /*
       * Old target was file
       */
      del_drive = to_cd->drive;
    }

    /* get rid of ent for old target */
    rc = nasd_edrfs_dir_remove_entry(to_cd, ditch_de);
    if (rc) {
      NASD_PANIC();
    }
  }
  else if ((to_cd == from_cd) && (from_ents_required == to_ents_required)) {
    /*
     * This works because they're in the same directory,
     * and we don't need to allocate more slots to hold
     * the name. Remember, in this case, the target does
     * not yet exist.
     */
    rc = nasd_edrfs_name_remove(from_cd, old_de, 1);
    if (rc) {
      NASD_PANIC();
    }
    new_de = old_de;
    nasd_edrfs_dir_dirty_page(to_cd, new_de->page);
    strcpy(new_de->name, in_to_path);
    new_de->data->name_len = to_len - 1; /* take 1 for NULL back off */
    rc = nasd_edrfs_name_hash_ins(to_cd, new_de);
    if (rc) {
      NASD_PANIC();
    }
  }
  else {
    /*
     * Target directory is not source directory, or target pages
     * different from source pages.
     *
     * Do it the hard way.
     */

#if DEBUG_RENAME_DETAIL > 0
    nasd_printf("rename between dirs\n");
    nasd_printf("rename old_de    0x%lx name \"%s\" id 0x%"
            NASD_ID_FMT "\n",
            old_de, old_de->name, old_de->data->nasdid);
    if (ditch_de) {
      nasd_printf("rename ditch_de 0x%lx name \"%s\" id 0x%"
              NASD_ID_FMT "\n",
              ditch_de, ditch_de->name, ditch_de->data->nasdid);
    }
#endif /* DEBUG_RENAME_DETAIL > 0 */

    /*
     * If we were actually fault-tolerant, we'd have a better way than
     * this. Here we first blow away the target if it exists, then
     * create the target ent, then blow away the source.
     */

    if (ditch_de) {
      del_nasdid = ditch_de->data->nasdid;
      if (new_cd) {
        /*
         * Old target was empty directory, new_cd is its core
         */
        new_cd->deleted = 1;
        NASD_EDRFS_UNLOCK_DIR(new_cd);
        while(new_cd->refcnt > 1) {
          NASD_WAIT_COND(new_cd->cond, nasd_edrfs_dircache_lock);
        }
        del_drive = new_cd->drive;
      }
      else {
        /*
         * Old target was file
         */
        del_drive = to_cd->drive;
      }

      /* get rid of ent for old target */
      rc = nasd_edrfs_dir_remove_entry(to_cd, ditch_de);
      if (rc) {
        NASD_PANIC();
      }
    }

    /*
     * Create target ent from scratch
     */
    flags = NASD_EDRFS_DIRD_BITS(old_de->data);
    rc = nasd_edrfs_dir_add_entry(to_cd, (char *)in_to_path,
      old_de->data->nasdid, flags, &new_de);
    if (rc) {
      *nasd_status = NASD_FAIL;
      goto done_rename2;
    }

#if DEBUG_RENAME_DETAIL > 0
    nasd_printf("remove entry for \"%s\"\n", old_de->name);
#endif /* DEBUG_RENAME_DETAIL > 0 */
    /* get rid of ent for original */
    rc = nasd_edrfs_dir_remove_entry(from_cd, old_de);
    if (rc) {
      NASD_PANIC();
    }
  }

  *nasd_status = NASD_SUCCESS;

done_rename2:
  if (new_cd) {
    NASD_EDRFS_UNLOCK_DIR(new_cd);
    nasd_edrfs_dir_release(new_cd);
    new_cd = NULL;
  }

  NASD_EDRFS_UNLOCK_DIR(from_cd);
  if (to_cd && (to_cd != from_cd)) {
    NASD_EDRFS_UNLOCK_DIR(to_cd);
  }

  if (del_drive) {
    rc = nasd_edrfs_do_remove(del_drive, del_nasdid);
    if (rc) {
      NASD_PANIC();
    }
  }

done_rename:
  if (from_cd) {
    nasd_edrfs_dir_release_post_attribute(from_cd, in_to_directory,
      post_to_attribute);
  }
  if (to_cd) {
    nasd_edrfs_dir_release_post_attribute(to_cd, in_to_directory,
      post_to_attribute);
  }
  NASD_EDRFS_UNLOCK_DIRCACHE();

really_done_rename:

  return;
}

void
nasd_edrfs_real_getstats(
  nasd_edrfs_opstats_t       *opstats,
  nasd_edrfs_cachestats_t    *cachestats,
  nasd_edrfs_rpc_opdepths_t  *opdepths,
  nasd_status_t              *nasd_status)
{
  if (cachestats) {
    *cachestats = nasd_edrfs_cache_stats;
  }
  if (opstats) {
    *opstats = nasd_edrfs_opstats;
  }
  if (opdepths) {
    *opdepths = nasd_edrfs_rpc_opdepths;
  }
  *nasd_status = NASD_SUCCESS;
}


/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
